const fs = require('fs');
const axios = require('axios');
const pluralize = require('pluralize');
const {Dictionary} = require('./utils/dictionary');
const {Threads} = require('./utils/threads');

const getApi = (gateway, token) => {
	if (!gateway || !token) {
		throw new Error('Gateway is not configured.');
	}
	const api = axios.create({
		baseURL: gateway
	})
	api.interceptors.request.use(
		config => {
			config.headers['Authorization'] = 'Bearer ' + token;
			return config;
		}
	)
	return api;
}

const buildQuery = (name, fields, variables) => {
	return `query ${
		variables ? `(${Object.entries(variables).map(([key, value]) => `$${key}:${value}`).join(',')})` : ''
	} { ${name}${
		variables ? `(${Object.keys(variables).map(name => `${name}:$${name}`).join(',')})` : ''
	}{ data{ ${fields.join(' ')} } meta { pagination { pageCount } } }}`;
};

/** Create promises for each collection **/
const getPromises = async (api, config) => {
	const promises = Object.keys(config)
		.reduce((obj, key) => {
			obj[key] = []
			return obj
		}, {})
	for (const [name, definition] of Object.entries(config)) {
		if (!definition?.fields) {
			throw new Error('Invalid configuration!');
		}
		const query = buildQuery(name, definition?.fields, {
			pagination: 'PaginationArg'
		})
		const response = await api.post('/', {
			query
		})
		const {pageCount} = response.data.data[name].meta.pagination
		let page = 1;
		while (page <= pageCount) {
			promises[name].push(api.post('/', {
				query,
				variables: {'pagination': {'page': page, 'pageSize': 10}}
			}));
			page++;
		}
	}
	return promises;
}

const unmaskField = (type, name, item) => {
	if (/^@+/.test(name)) {
		switch (name) {
			case '@uid':
				if (!item?.id) {
					throw new Error('Invalid object format!');
				}
				/**
				 * The unique identifier format
				 * supports 8-digits natural numbers only.
				 */
				return `${pluralize.singular(type)}@${item?.id.padStart(8, 0)}`;
			default:
				throw new Error(`Undefined mask: ${name}!`);
		}
	}

	if (!item[name]) {
		throw new Error(`Undefined key: ${name}!`);
	}
	return item[name];
};

/** Write data **/
const write = (path, name, data) => {
	fs.writeFile(`${fs.realpathSync(path)}/${name}.json`, JSON.stringify(data, null, 2), error => {
		if (error) {
			throw new Error(`Failed to write data to target dir: ${error?.message || error}`)
		}
		process.stdout.write('.')
	})
}

/** Download media **/
const download = (gateway, token, path, data) => {
	axios.get(`${gateway}${data.url}`, {
		responseType: 'arraybuffer',
		headers: {'Authorization': `Bearer ${token}`}
	})
		.then(res => {
			fs.writeFile(`${path}/${data.hash + data.ext}`, res.data, error => {
				if (error) {
					throw new Error('Failed to write media to target dir.')
				}
				process.stdout.write('+')
			});
		})
		.catch(error => {
			throw new Error(`Failed to download resource: ${error?.message || error}`)
		});
}

/** Fetch data **/
const fetchData = async (config, promises) => {
	let data = {};
	for (const key in promises) {
		await Promise.all(promises[key]).then((response) => {
			response = response.reduce((array, res) => {
				res.data.data[key]?.data.forEach(data => {

					/** Move attributes object 1 level up **/
					Object.entries(data.attributes).forEach(([key, value]) => {
						data[key] = value
					})
					delete data.attributes
					array.push(data)
				})
				return array
			}, [])
			data[key] = response
		});
	}
	return data;
}

module.exports = {
	fetch: async function (config, gateway, token) {
		const api = getApi(gateway, token);
		return getPromises(api, config).then(promises => fetchData(config, promises))
	},
	write: async function (response, config, gateway, token, dataPath, mediaPath) {

		const threads = new Threads((name, data) => {
			write(dataPath, name, data)
		});

		for (const [key, definition] of Object.entries(config)) {
			/** Performing the configured actions */
			definition?.actions.forEach(options => {
				if (!options) {
					throw new Error('Invalid configuration!');
				}
				if (Object.keys(response).length === 0) {
					throw new Error('Data is empty!');
				}
				switch (options?.method?.toLocaleString()) {
					case 'reference':
						if (!options?.key) {
							throw new Error('Invalid key!');
						}
						if (!options?.name) {
							throw new Error('Invalid name!');
						}
						if (options.type === 'id') {
							response[key].forEach(item => {
								item[options.key] = item[options.key].data ? unmaskField(options.key, options.name, item[options.key].data) : []
							})
						} else if (options.type === 'ids') {
							if (!options?.ref) {
								throw new Error('Invalid reference!');
							}
							response[key].forEach(item => {
								item[options.key] = item[options.key].data.length > 0 ?
									item[options.key].data.map((el) => {
										return unmaskField(options.ref, options.name, el)
									}) : []
							})
						} else if (options.type === 'media') {
							const [firstKey, secondKey] = options.name
							response[key].forEach(item => {
								item[options?.key] = item[options?.key]?.data?.attributes[firstKey] + item[options?.key]?.data?.attributes[secondKey]
							})
						}
						break;
					case 'extract':
						if (!options?.key) {
							throw new Error('Invalid key!');
						}

						if (!options?.value) {
							throw new Error('Invalid value!');
						}

						if (!options?.name) {
							throw new Error('Invalid name!');
						}
						const data = response[key].filter(item => item[options.key] === options.value)
						if (data.length > 0) {
							write(dataPath, options.name, options?.multiple ? data : data[0])
						}
						break;
					case 'collect':
						if (!options?.name) {
							throw new Error('Invalid name!');
						}
						response[key].forEach(item => {
							write(dataPath, unmaskField(key, options.name, item), item)
						})
						break;
					case 'compact':
						if (!options?.key) {
							throw new Error('Invalid key!');
						}
						if (!options?.value) {
							throw new Error('Invalid value!');
						}
						const compactor = new Dictionary((data) => {
							write(dataPath, key, data);
						});
						compactor.add(response[key], options.key, options.value)
						compactor.flush();
						break;
					case 'map':
						if (!options?.key) {
							throw new Error('Invalid key!');
						}

						if (!options?.value) {
							throw new Error('Invalid key!');
						}

						if (!options?.name) {
							throw new Error('Invalid name!');
						}

						response[key].forEach((item) => {
							threads.append(options.name, item[options.key], unmaskField(key, options.value, item));
						});
						break;
					case 'download':
						response[key].forEach((item) => {
							download(gateway, token, mediaPath, item);
						});
						break;
					default:
						throw new Error('Undefined method!');
				}
			})
		}
		threads.flush();
	}
}
