const fs = require('fs');
const Query = require('./queries/index');
const axios = require('axios');
const pluralize = require('pluralize');
const collections = ['articles', 'posts', 'categories', 'pages', 'authors', 'constants', 'uploadFiles'];

const getApiInstance = (url, token) => {
    if (!url || !token) {
        throw new Error('Gateway is not configured.');
    }
    const api = axios.create({
        baseURL: url
    })
    api.interceptors.request.use(
        config => {
            config.headers['Authorization'] = 'Bearer ' + token;
            return config;
        }
    )
    return api;
}

/** Get pagination meta data of each collection **/
const getPagination = async (api) => {
    const res = await api.post('/', {query: Query.pagination});
    return res.data.data;
}

/** Create promises for each collection based on pagination data **/
const getPromises = async (api, pagination) => {
    const promises = collections.reduce((obj, collection) => {
        obj[collection] = []
        return obj
    }, {})
    collections.forEach(key => {
        let page = 1;
        const pageSize = 10;
        const {pageCount} = pagination[key]?.meta?.pagination
        while (page <= pageCount) {
            promises[key].push(api.post('/', {
                query: Query[key],
                variables: {'pagination': {'page': page, 'pageSize': pageSize}}
            }));
            page++;
        }
    })
    return promises;
}

/** Get file name based on object id (@uid) **/
const getFileName = (key, id) => {
    return `${pluralize.singular(key)}@${id?.padStart(8, 0)}`
}

/** Mapping **/
const map = () => {

}

/** Fetch data **/
const fetchData = async (promises) => {
    let results = [];
    let routes = {};
    const keys = Object.keys(promises);
    for (let i = 0; i < keys.length; i++) {
        let key = keys[i];
        let data = await Promise.all(promises[key])
            .then((res) => {
                if (!res || !res.length) {
                    return [];
                }
                let items = [];
                res.forEach(r => {
                    items = [...items, ...r?.data?.data[key]?.data || []];
                })
                items.forEach((item, i) => {
                    /** Move attributes object 1 level up **/
                    const {attributes} = item
                    Object.entries(attributes).forEach(([key, value]) => {
                        item[key] = value
                    })
                    delete item.attributes

                    /** Move relation fields 1 level up **/
                    if (key === 'authors') {
                        const latest = item.latests.data.map(data => getFileName('post', data.id))
                        item.latests = latest
                        item.photo = item.photo.data?.attributes.hash + item.photo.data?.attributes.ext
                    }
                    if (key === 'categories') {
                        const related = item.related.data.map(data => getFileName('post', data.id))
                        item.related = related
                    }
                    if (key === 'posts') {
                        const category = getFileName('category', item.category.data.id)
                        const author = getFileName('author', item.author.data.id)
                        const resource = item.resource.data?.attributes.hash + item.resource.data?.attributes.ext
                        item.category = category
                        item.author = author
                        item.resource = resource || null

                        /** Remove featured from posts and create featured object**/
                        if (item.is_featured) {
                            results.push({key: 'featured', data: [item]})
                            // items.splice(i, 1)
                        }
                    }
                })
                /** Reduce constants into single object **/
                if (key === 'constants') {
                    const obj = items.reduce((obj, item) => {
                        obj[item.name] = item.value
                        return obj
                    }, {})
                    items = [obj]
                }

                /** Create routes object**/
                if (['pages', 'articles', 'posts', 'authors', 'categories'].includes(key)) {
                    routes = {
                        ...routes, ...
                            items.reduce((obj, item) => {
                                obj[item.url] = getFileName(key, item.id)
                                return obj
                            }, {})
                    }
                    results.push({key: 'routes', data: [routes]})
                }

                /** Create services object based on articles **/
                if (key === 'articles') {
                    const obj = items.reduce((obj, item) => {
                        obj[item.title] = item.url
                        return obj
                    }, {})
                    results.push({key: 'services', data: [obj]})
                }

                /** Formatted items **/
                return items;
            });
        results.push({key: key !== 'uploadFiles' ? key : 'resource', data});
    }
    return results;
}

/** Download resource **/
const download = (gatewayUrl, token, data, mediaPath) => {
    axios.get(`${gatewayUrl}${data.url}`, {
        responseType: 'arraybuffer',
        headers: {'Authorization': `Bearer ${token}`}
    })
        .then(res => {
            fs.writeFile(`${mediaPath}/${data.hash + data.ext}`, res.data, error => {
                if (error) {
                    throw new Error('Failed to write media to target dir.')
                } else {
                    process.stdout.write('+')
                }
            });
        })
        .catch(error => {
            throw new Error('Failed to download resource.')
        });
}

module.exports = {
    fetch: async function (url, token) {
        console.log('--Fetching...')
        const api = getApiInstance(url, token);
        return await getPagination(api)
            .then(pagination => getPromises(api, pagination))
            .then(promises => fetchData(promises));
    },
    write: async function (gatewayUrl, token, data, dataPath, mediaPath) {
        data.forEach(el => {
            el.data?.forEach(entry => {
                if (el.key === 'resource') {
                    download(gatewayUrl, token, entry, mediaPath)
                } else {
                    const fileName = !['featured', 'constants', 'routes', 'services'].includes(el.key) ? getFileName(el.key, entry.id) : el.key
                    fs.writeFile(`${fs.realpathSync(dataPath)}/${fileName}.json`, JSON.stringify(entry, null, 2), error => {
                        if (error) {
                            throw new Error('Failed to write data to target dir.')
                        } else {
                            process.stdout.write('.')
                        }
                    })
                }
            })
        })
    }
}